package pl.polgrabia.demos.crackingcodeinterview.t16x20;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.ArrayList;

public class RstTree {
    private static final Logger logger = LoggerFactory.getLogger(RstTree.class);
    private static final Logger log = LoggerFactory.getLogger(RstTree.class);
    private RstTreeNode rstTreeNode;

    public void add(String word) {
        assert word != null : "word must not be null";
        assert !word.isBlank() : "word must not be blank";
        assert word.matches("^[a-z]+$") : "word '" + word + "' must contain only small letters";
        if (rstTreeNode == null) {
            rstTreeNode = new RstTreeNode();
            rstTreeNode.setPrefix("");
            rstTreeNode.setValue(word);
        } else {
            addInternal(rstTreeNode, word);
        }
    }

    private static boolean addInternal(RstTreeNode node, String word) {
        if (word.equals(node.getValue())) {
            // already added
            return false;
        }

        assert word.startsWith(node.getPrefix()) : "word must start with prefix";
        int prefixLength = node.getPrefix().length();
        if (prefixLength >= word.length()) {
            String oldValue = node.getValue();
            node.setValue(word);
            boolean added = addInternal(node, oldValue);
            if (!added) {
                node.setValue(oldValue);
                return false;
            }

            return true;
        }
        char c = word.charAt(prefixLength);
        int offset = c - 'a';
        assert offset >= 0 && offset < RstTreeNode.EN_ALPHABET_LETTER_COUNT : "It's an offset for possible prefix child item - max alphabet";
        // logger.info("Word: {}, prefix: {}, index: {}, c: {}, offset: {}", word, node.getPrefix(), node.getPrefix().length(), c, offset);
        var childItem = node.getRstChildren().length <= offset ? null : node.getRstChildren()[offset];
        if (childItem == null) {
            childItem = new RstTreeNode();
            childItem.setPrefix(word.substring(0, prefixLength + 1));
            childItem.setValue(word);
            node.getRstChildren()[offset] = childItem;
            return true;
        } else {
            return addInternal(childItem, word);
        }
    }

    public boolean contains(String word) {
        assert word != null : "word must not be null";
        assert !word.isBlank() : "word must not be blank";
        assert word.matches("^[a-z]+$") : "word '" + word + "' must contain only small letters";

        return containsInternal(rstTreeNode, word);
    }

    private boolean containsInternal(RstTreeNode node, String word) {
        if (node == null) {
            return false;
        }

        logger.info("Node value: {}, prefix: {}, word: {}", node.getValue(), node.getPrefix(), word);

        assert word.startsWith(node.getPrefix()) : "word " + word + "must start with prefix: " + node.getPrefix();

        if (word.equals(node.getValue())) {
            return true;
        }

        char c = word.charAt(node.getPrefix().length());
        int offset = c - 'a';
        assert offset >= 0 && offset < RstTreeNode.EN_ALPHABET_LETTER_COUNT : "It's an offset for possible prefix child item - max alphabet";
        // logger.info("Word: {}, prefix: {}, index: {}, c: {}, offset: {}", word, node.getPrefix(), node.getPrefix().length(), c, offset);
        var childItem = node.getRstChildren().length <= offset ? null : node.getRstChildren()[offset];
        logger.info("Childitem: {}, c: {}, offset: {}", childItem, c, offset);
        return containsInternal(childItem, word);
    }
}
